#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import abc
from abc import ABC
from typing import Any, Dict, Iterable, List


class Table(ABC):
    """
    I'm a table in a relational database.
    """

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """
        The name of the table.
        :return: the name of this table
        """
        pass

    @abc.abstractmethod
    def fetch(self, primary_key: Dict[str, Any]) -> "RowSet":
        """
        Fetch rows with the associated primary key. The result will be a single row,
        but contained in a RowSet object.

        :param primary_key:  the key to look up by
        :return:  a RowSet with the row in it
        """
        pass

    @abc.abstractmethod
    def primary_key_columns(self) -> List[str]:
        """
        Return the primary key columns for this table.
        """
        pass

    @abc.abstractmethod
    def fetch_according_to(self, rows: "RowSet", columns: List[str]) -> "RowSet":
        """
        Fetch rows according to the supplied rows and columns

        :param rows:    rows to mine for the WHERE clause
        :param columns: columns to consider in the generated WHERE clause
        :return: RowSet for the rows in this table
        """
        pass

    @abc.abstractmethod
    def relations(self) -> Iterable["Relationship"]:
        """
        Return the relationships that might pertain to the rows we have here.

        :return:
        """
        pass


class RowWriter:
    def write_rows(self, table: Table, rows: List[Dict]):
        raise NotImplementedError

    def close(self):
        """
        By default there's nothing to do here, but subclasses can override this
        if the need to catch a signal that we're done writing.
        """
        pass


class RowSet(ABC):
    """
    I'm a set of rows from a SELECT * or TABLE query, run against a relational database.
    """

    @property
    @abc.abstractmethod
    def table(self) -> Table:
        pass

    @abc.abstractmethod
    def relations(self) -> Iterable["Relationship"]:
        """
        Return the relationships that might pertain to the rows we have here.

        :return:
        """
        pass

    @abc.abstractmethod
    def write_to(self, writer: RowWriter):
        """
        Write this rowset to a particular row writer
        :param writer: write to communicate with
        """
        pass

    @abc.abstractmethod
    def __iter__(self) -> Iterable[Dict]:
        pass

    @abc.abstractmethod
    def __len__(self) -> int:
        pass


class Relationship(ABC):
    """
    I'm a relationship between two tables, typically a foreign key relationship.
    """

    @property
    @abc.abstractmethod
    def source_table(self) -> Table:
        """
        The table with the foreign key. Consider the relationship
        between execution_blocks(project_code) -> projects(project_code). The
        source table is execution_blocks.
        :return:
        """
        pass

    @property
    @abc.abstractmethod
    def destination_table(self) -> Table:
        """
        The table with the foreign key. Consider the relationship
        between execution_blocks(project_code) -> projects(project_code). The
        destination table is projects.
        :return:
        """
        pass

    @abc.abstractmethod
    def fetch_related_to(self, rows: RowSet) -> RowSet:
        """
        Return rows related to the supplied rows based on this relationship.
        For instance, if this is a relationship from projects to science products,
        you can supply a RowSet of projects to this relationship, and receive a
        new RowSet of science products.

        :param rows:  source rows
        :return:      rows from the destination
        """
        pass
